from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from clickhouse_query.models import Value


class UIDGenerator:
    def __init__(self, prefix=None):
        if prefix is None:
            prefix = ""
        self.prefix = prefix
        self.prev_num = 0

    def get(self):
        self.prev_num += 1
        return self.prefix + str(self.prev_num)


def get_sql(arg, uid_generator: UIDGenerator | None = None):
    from clickhouse_query import mixins

    if uid_generator is None:
        uid_generator = UIDGenerator(prefix="__U_")

    sql, sql_params = arg.__sql__(uid_generator=uid_generator)

    if isinstance(arg, mixins.ASMixin):
        sql += arg.__asmixin__(uid_generator=uid_generator)

    return sql, sql_params


def _extract_condition(item: str, value: "Value"):
    from clickhouse_query import functions, models
    from clickhouse_query.config import INLINE_CONDITIONS

    *_field, op = item.split("__")

    field = models.F(_field)

    if op in INLINE_CONDITIONS:
        return INLINE_CONDITIONS[op](field, value)

    field = models.F(_field + [op])
    return functions.equals(field, value)


def _apply_operator(field, op):
    from clickhouse_query.config import INLINE_FUNCTIONS

    if op[0] == "_":  # nested operator
        return field._arg_extend("." + op[1:])
    elif op in INLINE_FUNCTIONS:
        return INLINE_FUNCTIONS[op](field)

    return field._arg_extend("__" + op)


def get_expression(v, str_is_field=False):
    from clickhouse_query import models

    if str_is_field and isinstance(v, str):
        return models.F(v)

    if hasattr(v, "__sql__"):
        return v

    return models.Value(v)


def is_iterable(obj):
    try:
        iter(obj)
        return True
    except TypeError:
        return False



from clickhouse_query import mixins
from clickhouse_query.functions import Function, functions
from clickhouse_query.models import NULL, F, Q, QuerySet, Subquery, Value
from clickhouse_query.utils import get_sql

__all__ = [
    "mixins",
    "Function",
    "functions",
    "NULL",
    "F",
    "Q",
    "QuerySet",
    "Subquery",
    "Value",
    "get_sql",
]



from clickhouse_query import mixins, utils


class Function(mixins.ASMixin, mixins.ArithmeticMixin):
    def __init__(self, function_name):
        self.args_list: list[list] = []
        self.function_name = function_name

    def __call__(self, *args):
        self.args_list.append(args)
        return self

    def get_function_name(self) -> str:
        return self.function_name

    def __sql__(self, uid_generator):
        sql_params = {}
        sql = self.get_function_name()
        for args in self.args_list:
            args_sqls_list = []
            for item in args:
                expression = utils.get_expression(item, str_is_field=True)
                item_sql, item_sql_params = utils.get_sql(expression, uid_generator=uid_generator)
                args_sqls_list.append(item_sql)
                sql_params.update(item_sql_params)
            sql += "({args})".format(args=", ".join(args_sqls_list))
        return sql, sql_params


class _FunctionBuilder:
    def __getattr__(self, item: str) -> Function:
        if item[-1] == "_":
            item = item[:-1]
        return Function(item)


functions = _FunctionBuilder()



from clickhouse_query import functions, models

__all__ = ["INLINE_CONDITIONS", "INLINE_FUNCTIONS"]


def _exact(field, value):
    if value is None:
        return functions.isNull(field)
    else:
        return functions.equals(field, value)


def _iexact(field, value):
    if value is None:
        return functions.isNull(field)
    else:
        return functions.iLike(field, value)


def _contains(field, value):
    return functions.like(field, functions.concat(models.Value("%"), value, models.Value("%")))


def _icontains(field, value):
    return functions.iLike(field, functions.concat(models.Value("%"), value, models.Value("%")))


def _isnull(field, value):
    if value.arg is True:
        return functions.isNull(field)
    elif value.arg is False:
        return functions.isNotNull(field)
    else:
        raise ValueError("isnull value should be True or False")


INLINE_CONDITIONS = {
    "exact": _exact,
    "iexact": _iexact,
    "contains": _contains,
    "icontains": _icontains,
    "in": functions.in_,
    "notin": functions.notIn,
    "gt": functions.greater,
    "gte": functions.greaterOrEquals,
    "lt": functions.less,
    "lte": functions.lessOrEquals,
    "startswith": lambda field, value: functions.like(field, functions.concat(value, models.Value("%"))),
    "istartswith": lambda field, value: functions.iLike(field, functions.concat(value, models.Value("%"))),
    "endswith": lambda field, value: functions.like(field, functions.concat(models.Value("%"), value)),
    "iendswith": lambda field, value: functions.iLike(field, functions.concat(models.Value("%"), value)),
    "isnull": _isnull,
}


INLINE_FUNCTIONS = {
    "year": functions.toYear,
    "month": functions.toMonth,
    "day": functions.toDayOfYear,
    "week": functions.toWeek,
    "week_day": functions.toDayOfWeek,
    "quarter": functions.toQuarter,
    "time": functions.toTime,
    "hour": functions.toHour,
    "minute": functions.toMinute,
    "second": functions.toSecond,
}



import copy


class ASMixin:
    _AS_FORCE = False

    def as_(self, as_):
        self = copy.deepcopy(self)
        self._as = as_
        return self

    def _get_as(self, uid_generator):
        as_ = getattr(self, "_as", None)
        if as_ is None and self._AS_FORCE:
            if not hasattr(self, "uid_generator_id") or id(uid_generator) != self.uid_generator_id:
                self.uid_generator_id = id(uid_generator)
                self.uid = uid_generator.get()
            return self.uid
        return as_

    def has_as(self, uid_generator):
        return self._get_as(uid_generator=uid_generator) is not None

    def __asmixin__(self, uid_generator):
        as_ = self._get_as(uid_generator=uid_generator)
        if as_ is not None:
            return " AS {as_}".format(as_=as_)
        return ""


class ArithmeticMixin:
    def __add__(self, other):
        from clickhouse_query import functions

        return functions.plus(self, other)

    def __sub__(self, other):
        from clickhouse_query import functions

        return functions.minus(self, other)

    def __mul__(self, other):
        from clickhouse_query import functions

        return functions.multiply(self, other)

    def __truediv__(self, other):
        from clickhouse_query import functions

        return functions.divide(self, other)

    def __floordiv__(self, other):
        from clickhouse_query import functions

        # TODO: cast self to Float64
        return functions.intDiv(self, other)

    def __mod__(self, other):
        from clickhouse_query import functions

        return functions.modulo(self, other)

    def __neg__(self):
        from clickhouse_query import functions

        return functions.negate(self)



import copy
import datetime
from zoneinfo import ZoneInfo

from clickhouse_query import Function, mixins, utils


class QuerySet:
    def __init__(self, _from=None, inplace=False, clickhouse_client=None):
        self._from = None
        self._select = []
        self._distinct = None
        self._prewhere = []
        self._where = []
        self._group_by = []
        self._having = []
        self._order_by = []
        self._limit = None
        self._limit_by = None

        self.inplace = inplace
        self.clickhouse_client = clickhouse_client

        self.enable_inplace().from_(_from).disable_inplace()

    def _clone(self):
        if not self.inplace:
            self = copy.deepcopy(self)
        return self

    def disable_inplace(self):
        self.inplace = False
        return self

    def enable_inplace(self):
        self.inplace = True
        return self

    def select(self, *args, **kwargs):
        self = self._clone()

        if args and args[0] is None:
            self._select = []
            return self

        self._select = []
        for item in args:
            expression = utils.get_expression(item, str_is_field=True)
            self._select.append(expression)
        for as_, item in kwargs.items():
            expression = utils.get_expression(item, str_is_field=True)
            expression = expression.as_(as_)
            self._select.append(expression)
        return self

    def distinct(self, *args):
        self = self._clone()

        if args and args[0] is None:
            self._distinct = None
            return self

        self._distinct = []
        for item in args:
            expression = utils.get_expression(item, str_is_field=True)
            self._distinct.append(expression)
        return self

    def from_(self, from_):
        self = self._clone()

        if from_ is None:
            self._from = None
            return self

        self._from = utils.get_expression(from_, str_is_field=True)
        return self

    def prewhere(self, *args, **kwargs):
        self = self._clone()

        if args and args[0] is None:
            self._prewhere = []
            return self

        prewhere_dict_expr = {k: utils.get_expression(v) for k, v in kwargs.items()}
        prewhere_condition_list = [utils._extract_condition(k, v) for k, v in prewhere_dict_expr.items()]
        self._prewhere += list(args) + prewhere_condition_list
        return self

    def where(self, *args, **kwargs):
        self = self._clone()

        if args and args[0] is None:
            self._where = []
            return self

        where_dict_expr = {k: utils.get_expression(v) for k, v in kwargs.items()}
        where_condition_list = [utils._extract_condition(k, v) for k, v in where_dict_expr.items()]
        self._where += list(args) + where_condition_list
        return self

    def group_by(self, *args):
        self = self._clone()

        if args and args[0] is None:
            self._group_by = []
            return self

        self._group_by = []
        for item in args:
            expression = utils.get_expression(item, str_is_field=True)
            self._group_by.append(expression)
        return self

    def having(self, *args, **kwargs):
        self = self._clone()

        if args and args[0] is None:
            self._having = []
            return self

        having_dict_expr = {k: utils.get_expression(v) for k, v in kwargs.items()}
        having_condition_list = [utils._extract_condition(k, v) for k, v in having_dict_expr.items()]
        self._having += list(args) + having_condition_list
        return self

    def order_by(self, *args):
        self = self._clone()

        if args and args[0] is None:
            self._order_by = []
            return self

        self._order_by = []
        for item in args:
            if not isinstance(item, OrderBy):
                item = OrderBy.get_from_str(item)
            self._order_by.append(item)
        return self

    def limit(self, limit, *, offset=None):
        self = self._clone()

        if limit is None and offset is None:
            self._limit = None
            return self

        if limit is not None:
            limit = utils.get_expression(limit)
        if offset is not None:
            offset = utils.get_expression(offset)

        self._limit = (limit, offset)
        return self

    def limit_by(self, limit, *by, offset=None):
        self = self._clone()

        if limit is None:
            self._limit_by = None
            return self

        limit = utils.get_expression(limit)
        if offset is not None:
            offset = utils.get_expression(offset)

        _by = []
        for item in by:
            expression = utils.get_expression(item, str_is_field=True)
            _by.append(expression)

        self._limit_by = (limit, offset, _by)
        return self

    def _get_select_sql(self, uid_generator):
        if not self._select:
            return " *", {}

        sql_params = {}
        sqls = []
        for expression in self._select:
            sql, params = utils.get_sql(expression, uid_generator=uid_generator)
            sqls.append(sql)
            sql_params.update(params)

        select_sql = " {}".format(", ".join(sqls))
        return select_sql, sql_params

    def _get_distinct_sql(self, uid_generator):
        if self._distinct is None:
            return "", {}
        distinct_sql = " DISTINCT"
        sql_params = {}
        if self._distinct:
            sqls = []
            for expression in self._distinct:
                sql, params = utils.get_sql(expression, uid_generator=uid_generator)
                sqls.append(sql)
                sql_params.update(params)

            distinct_sql += " ON ({})".format(", ".join(sqls))
        return distinct_sql, sql_params

    def _get_from_sql(self, uid_generator):
        if self._from is None:
            return "", {}

        expression = self._from
        sql, sql_params = utils.get_sql(expression, uid_generator=uid_generator)
        from_sql = " FROM {}".format(sql)
        return from_sql, sql_params

    def _get_prewhere_sql(self, uid_generator):
        if not self._prewhere:
            return "", {}

        sql, sql_params = utils.get_sql(Q(*self._prewhere), uid_generator=uid_generator)
        prewhere_sql = " PREWHERE {}".format(sql)
        return prewhere_sql, sql_params

    def _get_where_sql(self, uid_generator):
        if not self._where:
            return "", {}

        sql, sql_params = utils.get_sql(Q(*self._where), uid_generator=uid_generator)
        where_sql = " WHERE {}".format(sql)
        return where_sql, sql_params

    def _get_group_by_sql(self, uid_generator):
        if not self._group_by:
            return "", {}

        sql_params = {}
        sqls = []
        for expression in self._group_by:
            sql, params = utils.get_sql(expression, uid_generator=uid_generator)
            sqls.append(sql)
            sql_params.update(params)

        group_by_sql = " GROUP BY {}".format(", ".join(sqls))
        return group_by_sql, sql_params

    def _get_having_sql(self, uid_generator):
        if not self._having:
            return "", {}

        sql, sql_params = utils.get_sql(Q(*self._having), uid_generator=uid_generator)
        having_sql = " HAVING {}".format(sql)
        return having_sql, sql_params

    def _get_order_by_sql(self, uid_generator):
        if not self._order_by:
            return "", {}

        sql_params = {}
        sqls = []
        for expression in self._order_by:
            sql, params = utils.get_sql(expression, uid_generator=uid_generator)
            sqls.append(sql)
            sql_params.update(params)

        order_by_sql = " ORDER BY {}".format(", ".join(sqls))
        return order_by_sql, sql_params

    def _get_limit_sql(self, uid_generator):
        if self._limit is None or self._limit == (None, None):
            return "", {}

        limit_expr, offset_expr = self._limit
        sql_params = {}

        if limit_expr is not None:
            _limit_sql, limit_params = utils.get_sql(limit_expr, uid_generator=uid_generator)
            limit_sql = " LIMIT {}".format(_limit_sql)
            sql_params.update(limit_params)

        if offset_expr is not None:
            _offset_sql, offset_params = utils.get_sql(offset_expr, uid_generator=uid_generator)
            limit_sql += " OFFSET {}".format(_offset_sql)
            sql_params.update(offset_params)

        return limit_sql, sql_params

    def _get_limit_by_sql(self, uid_generator):
        if self._limit_by is None:
            return "", {}

        limit_expr, offset_expr, by_exprs = self._limit_by
        sql_params = {}

        _limit_sql, limit_params = utils.get_sql(limit_expr, uid_generator=uid_generator)
        limit_by_sql = " LIMIT {}".format(_limit_sql)
        sql_params.update(limit_params)

        if offset_expr is not None:
            _offset_sql, offset_params = utils.get_sql(offset_expr, uid_generator=uid_generator)
            limit_by_sql += " OFFSET {}".format(_offset_sql)
            sql_params.update(offset_params)

        _by_sqls = []
        for expression in by_exprs:
            sql, params = utils.get_sql(expression, uid_generator=uid_generator)
            _by_sqls.append(sql)
            sql_params.update(params)
        limit_by_sql += " BY {}".format(", ".join(_by_sqls))

        return limit_by_sql, sql_params

    def __sql__(self, *, uid_generator):
        distinct_sql, distinct_params = self._get_distinct_sql(uid_generator=uid_generator)
        select_sql, select_params = self._get_select_sql(uid_generator=uid_generator)
        from_sql, from_params = self._get_from_sql(uid_generator=uid_generator)
        prewhere_sql, prewhere_params = self._get_prewhere_sql(uid_generator=uid_generator)
        where_sql, where_params = self._get_where_sql(uid_generator=uid_generator)
        group_by_sql, group_by_params = self._get_group_by_sql(uid_generator=uid_generator)
        having_sql, having_params = self._get_having_sql(uid_generator=uid_generator)
        order_by_sql, order_by_params = self._get_order_by_sql(uid_generator=uid_generator)
        limit_by_sql, limit_by_params = self._get_limit_by_sql(uid_generator=uid_generator)
        limit_sql, limit_params = self._get_limit_sql(uid_generator=uid_generator)

        sql_params = {}
        sql_params.update(distinct_params)
        sql_params.update(select_params)
        sql_params.update(from_params)
        sql_params.update(prewhere_params)
        sql_params.update(where_params)
        sql_params.update(group_by_params)
        sql_params.update(having_params)
        sql_params.update(order_by_params)
        sql_params.update(limit_by_params)
        sql_params.update(limit_params)

        str_format = (
            "SELECT{distinct}{select}{from_}{prewhere}{where}{group_by}{having}{order_by}{limit_by}{limit}"
        )
        sql = str_format.format(
            distinct=distinct_sql,
            select=select_sql,
            from_=from_sql,
            prewhere=prewhere_sql,
            where=where_sql,
            group_by=group_by_sql,
            having=having_sql,
            order_by=order_by_sql,
            limit_by=limit_by_sql,
            limit=limit_sql,
        )
        return sql, sql_params

    def execute(self, clickhouse_client=None):
        if clickhouse_client is None:
            clickhouse_client = self.clickhouse_client

        if callable(clickhouse_client):
            clickhouse_client = clickhouse_client()

        sql, sql_params = utils.get_sql(self)

        values_list, column__type = clickhouse_client.execute(sql, params=sql_params, with_column_types=True)

        data = []
        for values in values_list:
            data.append({column: value for (column, type), value in zip(column__type, values)})  # noqa: B905

        return data


class Subquery(mixins.ASMixin):
    def __init__(self, inner_queryset):
        self._inner_queryset = inner_queryset

    def __sql__(self, *, uid_generator):
        inner_queryset_sql, sql_params = utils.get_sql(self._inner_queryset, uid_generator=uid_generator)
        sql = "({inner_queryset})".format(inner_queryset=inner_queryset_sql)
        return sql, sql_params


class _Null:
    def __sql__(self, *, uid_generator):
        return "NULL", {}


NULL = _Null()


class _F:
    def __init__(self, arg):
        self.arg = arg

    def __sql__(self, *, uid_generator):
        return str(self.arg), {}

    def _arg_extend(self, arg):  # TODO: rename
        return _F(self.arg + arg)


class F(mixins.ASMixin, mixins.ArithmeticMixin):
    def __init__(self, arg):
        self.arg = arg.split("__") if isinstance(arg, str) else arg

    def __sql__(self, *, uid_generator):
        field, *operators = self.arg
        field = _F(field)
        for op in operators:
            field = utils._apply_operator(field, op)

        sql, sql_params = utils.get_sql(field, uid_generator=uid_generator)
        return sql, sql_params


def _get_value_sql(arg):
    if isinstance(arg, str):
        return "%({uid})s", arg

    if isinstance(arg, (int, float)):
        return "%({uid})f", arg

    if isinstance(arg, datetime.datetime):
        return "%({uid})s", arg.astimezone(ZoneInfo("UTC")).strftime("%Y-%m-%d %H:%M:%S.%f")

    if utils.is_iterable(arg):
        new_arg = tuple(_get_value_sql(a)[1] for a in arg)
        return "%({uid})s", new_arg

    raise Exception("Value not Valid", arg)


class Value(mixins.ASMixin, mixins.ArithmeticMixin):
    def __init__(self, arg):
        self.arg = arg

    def __sql__(self, *, uid_generator):
        uid = uid_generator.get()
        value_sql, value_param = _get_value_sql(self.arg)
        return value_sql.format(uid=uid), {uid: value_param}


class OrderBy:
    def __init__(self, expr, asc=True):
        self.expr = expr
        self.asc = asc

    def __sql__(self, *, uid_generator):
        sql, sql_params = utils.get_sql(self.expr, uid_generator=uid_generator)
        return "{} {}".format(sql, "ASC" if self.asc else "DESC"), sql_params

    @classmethod
    def get_from_str(cls, arg):
        if arg.startswith("-"):
            expr = utils.get_expression(arg[1:], str_is_field=True)
            asc = False
        else:
            expr = utils.get_expression(arg, str_is_field=True)
            asc = True
        return cls(expr, asc=asc)


class Q(Function):
    def __init__(self, *args, _connector="and", **kwargs) -> None:
        args = list(args)

        if len(args) + len(kwargs) == 0:
            raise ValueError("At least one argument is required")

        if len(args) + len(kwargs) == 1:
            _connector = ""

        for item, value in kwargs.items():
            expression = utils.get_expression(value)
            condition = utils._extract_condition(item, expression)
            args.append(condition)

        super().__init__(_connector)
        self(*args)



import clickhouse_query as ch


def test_select():
    q = ch.QuerySet().select("a")
    assert ch.get_sql(q) == ("SELECT a", {})
    q = ch.QuerySet().select(a="a")
    assert ch.get_sql(q) == ("SELECT a AS a", {})

    q = ch.QuerySet().select(1)
    assert ch.get_sql(q) == ("SELECT %(__U_1)f", {"__U_1": 1})
    q = ch.QuerySet().select(a=1)
    assert ch.get_sql(q) == ("SELECT %(__U_1)f AS a", {"__U_1": 1})

    q = ch.QuerySet().select(ch.Value("a"))
    assert ch.get_sql(q) == ("SELECT %(__U_1)s", {"__U_1": "a"})
    q = ch.QuerySet().select(a=ch.Value("a"))
    assert ch.get_sql(q) == ("SELECT %(__U_1)s AS a", {"__U_1": "a"})

    q = ch.QuerySet().select([1, 2, 3])
    assert ch.get_sql(q) == ("SELECT %(__U_1)s", {"__U_1": (1, 2, 3)})
    q = ch.QuerySet().select(a=[1, 2, 3])
    assert ch.get_sql(q) == ("SELECT %(__U_1)s AS a", {"__U_1": (1, 2, 3)})


def test_from():
    q = ch.QuerySet().from_("table1")
    assert ch.get_sql(q) == ("SELECT * FROM table1", {})

    sub_q = ch.QuerySet().from_("table1")
    q = ch.QuerySet().from_(ch.Subquery(sub_q))
    assert ch.get_sql(q) == ("SELECT * FROM (SELECT * FROM table1)", {})


def test_distinct():
    q = ch.QuerySet().distinct()
    assert ch.get_sql(q) == ("SELECT DISTINCT *", {})

    q = ch.QuerySet().select("a").distinct()
    assert ch.get_sql(q) == ("SELECT DISTINCT a", {})

    q = ch.QuerySet().select("a").distinct("b")
    assert ch.get_sql(q) == ("SELECT DISTINCT ON (b) a", {})


def test_prewhere():
    q = ch.QuerySet().from_("t1").prewhere(a="b")
    assert ch.get_sql(q) == ("SELECT * FROM t1 PREWHERE (equals(a, %(__U_1)s))", {"__U_1": "b"})

    q = ch.QuerySet().from_("t1").prewhere(ch.functions.equals("a", ch.Value("b")))
    assert ch.get_sql(q) == ("SELECT * FROM t1 PREWHERE (equals(a, %(__U_1)s))", {"__U_1": "b"})

    q = ch.QuerySet().from_("t1").prewhere(a="b", c="d")
    assert ch.get_sql(q) == (
        "SELECT * FROM t1 PREWHERE and(equals(a, %(__U_1)s), equals(c, %(__U_2)s))",
        {"__U_1": "b", "__U_2": "d"},
    )


def test_functions():
    f = ch.functions.func(1, 2)
    assert ch.get_sql(f) == ("func(%(__U_1)f, %(__U_2)f)", {"__U_1": 1, "__U_2": 2})

    f = ch.functions.func(1, 2).as_("f1")
    assert ch.get_sql(f) == ("func(%(__U_1)f, %(__U_2)f) AS f1", {"__U_1": 1, "__U_2": 2})

    f = ch.functions.func(1, 2) + 1
    assert ch.get_sql(f) == (
        "plus(func(%(__U_1)f, %(__U_2)f), %(__U_3)f)",
        {"__U_1": 1, "__U_2": 2, "__U_3": 1},
    )

    f = ch.functions.sumIf("a", "b")
    assert ch.get_sql(f) == ("sumIf(a, b)", {})


def test_group_by():
    q = ch.QuerySet().select(key="col1", a=ch.functions.count()).from_("t1").group_by("key")
    assert ch.get_sql(q) == ("SELECT col1 AS key, count() AS a FROM t1 GROUP BY key", {})

    q = ch.QuerySet().select(key="col1", a=ch.functions.count()).from_("t1").group_by("key", "key2")
    assert ch.get_sql(q) == ("SELECT col1 AS key, count() AS a FROM t1 GROUP BY key, key2", {})


def test_order_by():
    q = ch.QuerySet().order_by("col1")
    assert ch.get_sql(q) == ("SELECT * ORDER BY col1 ASC", {})

    q = ch.QuerySet().order_by("-col1")
    assert ch.get_sql(q) == ("SELECT * ORDER BY col1 DESC", {})

    q = ch.QuerySet().order_by("col1", "-col2")
    assert ch.get_sql(q) == ("SELECT * ORDER BY col1 ASC, col2 DESC", {})


def test_q():
    f = ch.Q(a=1, b=2)
    assert ch.get_sql(f) == ("and(equals(a, %(__U_1)f), equals(b, %(__U_2)f))", {"__U_1": 1, "__U_2": 2})

    f = ch.Q(a=1, b=2, _connector="or")
    assert ch.get_sql(f) == ("or(equals(a, %(__U_1)f), equals(b, %(__U_2)f))", {"__U_1": 1, "__U_2": 2})

    f = ch.Q(a=1, b=2, _connector="my_connector")
    assert ch.get_sql(f) == (
        "my_connector(equals(a, %(__U_1)f), equals(b, %(__U_2)f))",
        {"__U_1": 1, "__U_2": 2},
    )


def test_limit():
    q = ch.QuerySet().limit(2)
    assert ch.get_sql(q) == ("SELECT * LIMIT %(__U_1)f", {"__U_1": 2})

    q = ch.QuerySet().limit(5, offset=10)
    assert ch.get_sql(q) == ("SELECT * LIMIT %(__U_1)f OFFSET %(__U_2)f", {"__U_1": 5, "__U_2": 10})

    q = ch.QuerySet().limit(None)
    assert ch.get_sql(q) == ("SELECT *", {})


def test_limit_by():
    q = ch.QuerySet().limit_by(2, "col1")
    assert ch.get_sql(q) == ("SELECT * LIMIT %(__U_1)f BY col1", {"__U_1": 2})

    q = ch.QuerySet().limit_by(2, "col1", "col2")
    assert ch.get_sql(q) == ("SELECT * LIMIT %(__U_1)f BY col1, col2", {"__U_1": 2})


def test_custom_inline_condition(mocker):
    mocker.patch("clickhouse_query.config.INLINE_CONDITIONS", {"something": ch.functions.something})

    q = ch.QuerySet().where(a__something=1)
    assert ch.get_sql(q) == ("SELECT * WHERE (something(a, %(__U_1)f))", {"__U_1": 1})


def test_custom_inline_function(mocker):
    mocker.patch("clickhouse_query.config.INLINE_FUNCTIONS", {"something": ch.functions.something})

    q = ch.QuerySet().where(a__something__lte=1)
    assert ch.get_sql(q) == ("SELECT * WHERE (lessOrEquals(something(a), %(__U_1)f))", {"__U_1": 1})



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="clickhouse-query",
    version="0.1.1",
    author="Mehrshad Hosseini",
    author_email="mehrh8@gmail.com",
    description="clickhouse query",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/mehrh8/clickhouse-query",
    project_urls={
        "Bug Tracker": "https://github.com/mehrh8/clickhouse-query/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    packages=["clickhouse_query"],
    python_requires=">=3.6",
)
